"""
"""
from typing import Dict, List, Optional
from copy import deepcopy
from .decorators import register_var

default_fn_dict = {
    "type": "choice",
    "argument_fn": {},
    "default": None,
}
default_val_dict = {
    "type": "val",
    "dtype": None,
    "default": None,
}

class VariableClass(object):
    var_name: str = ""
    default: Optional[str] = None

class RegisteringChoiceType(type):
    def __init__(cls, name, bases, attrs):
        cls.arguments: List[str] = []
        cls.variables: Dict[str, dict] = {}
        cls.variable_shown_name: Dict[str, Dict[str, str]] = {}
        if hasattr(cls, 'var_name') and cls.var_name is not None:
            name = cls.var_name
        elif not hasattr(cls, 'var_name'):
            # shouldn't happend
            cls.var_name = name

        for key, val in attrs.items():
            props = getattr(val, 'registers', None)
            if props is None:
                continue
            for prop in props:
                var_name = name if not prop['var_name'] else prop['var_name']
                argument = key if not prop['argument'] else prop['argument']
                shown_name = key if not prop['shown_name'] else prop['shown_name']
                cls.variables.setdefault(var_name, deepcopy(default_fn_dict))["argument_fn"][argument] = val.__func__
                cls.variables[var_name].setdefault('required_vars', {})[argument] = prop['required_vars'] if 'required_vars' in prop else None
                cls.variables[var_name].setdefault('cache_dirs', {})[argument] = prop['cache_dir'] if 'cache_dir' in prop else None
                cls.arguments.append(argument)
                cls.variable_shown_name.setdefault(var_name, dict())[argument] = shown_name

class ParameterAlreadyRanError(Exception):
    def __init__(self, message="", errors=""):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors

class VariableValueNotSetError(Exception):
    def __init__(self, message="", errors=""):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors

class VariableNotRegisteredError(Exception):
    def __init__(self, message="", errors=""):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors



import signal
import sys
import os
import urllib.parse
import tempfile
import json
import copy
from typing import Dict, Tuple, List, Any, Callable, AnyStr
import logging

import requests
from mkdir_p import mkdir_p
import joblib

from ..base import ParameterAlreadyRanError
from ..auto_var import AutoVar

_logger = logging.getLogger(__name__)

def get_ext(file_format: str) -> str:
    if file_format == 'json':
        return 'json'
    elif file_format == 'pickle':
        return 'pkl'
    else:
        raise ValueError(f"Not supported file format {file_format}")

def default_get_file_name(auto_var: AutoVar) -> str:
    file_name_list: List[str] = []
    for k, v in sorted(auto_var.var_value.items()):
        if k != 'git_hash':
            file_name_list.append(f"{str(v)}")
    file_name: str = '-'.join(file_name_list)
    return file_name

def check_result_file_exist(auto_var, get_name_fn=None):
    if get_name_fn is None:
        get_name_fn = default_get_file_name
    unique_name = get_name_fn(auto_var)
    unique_name = f'{unique_name}.{get_ext(auto_var.settings["file_format"])}'
    base_dir = auto_var.settings['result_file_dir']
    file_path = os.path.join(base_dir, unique_name)
    if os.path.exists(file_path):
        _logger.warning(f"{file_path} exists")
        raise ParameterAlreadyRanError("%s exists" % file_path)

def create_placeholder_file(auto_var, get_name_fn=None):
    if get_name_fn is None:
        get_name_fn = default_get_file_name
    unique_name = get_name_fn(auto_var)
    unique_name = f'{unique_name}.{get_ext(auto_var.settings["file_format"])}'
    base_dir = auto_var.settings['result_file_dir']
    file_path = os.path.join(base_dir, unique_name)
    with open(file_path, "w") as f:
        f.write("placeholder, program still running")
        
def remove_placeholder_if_error(auto_var, ret, get_name_fn=None):
    if ret is None:
        unique_name = get_name_fn(auto_var)
        unique_name = f'{unique_name}.{get_ext(auto_var.settings["file_format"])}'
        base_dir = auto_var.settings['result_file_dir']
        file_path = os.path.join(base_dir, unique_name)
        _logger.warning(f"removing {file_path} ...")
        os.unlink(file_path)

def save_result_to_file(auto_var, ret, get_name_fn=None):
    if get_name_fn is None:
        get_name_fn = default_get_file_name
    base_dir = auto_var.settings['result_file_dir']
    file_format = auto_var.settings["file_format"]
    unique_name = get_name_fn(auto_var)
    output_file = os.path.join(base_dir, f'{unique_name}.{get_ext(file_format)}')
    if file_format == 'json':
        with open(output_file, "w") as f:
            json.dump(ret, f)
    elif file_format == 'pickle':
        with open(output_file, "wb") as f:
            joblib.dump(ret, f)
    else:
        raise ValueError(f"Not supported file format {file_format}")
    _logger.info("Finish writing to file %s", output_file)

def save_parameter_to_file(auto_var, get_name_fn=None):
    if get_name_fn is None:
        get_name_fn = default_get_file_name

    unique_name = get_name_fn(auto_var) + "_param"
    output_file = '%s.json' % unique_name
    with open(output_file, "w") as f:
        json.dump(auto_var.var_value, f)
    _logger.info("Finish writing to file %s", output_file)

def upload_result(auto_var, ret):
    #files = {'file_field': json.dumps(ret)}
    if auto_var.parameter_id is None:
        raise ValueError
    url = urllib.parse.urljoin(auto_var.settings['server_url'], 'upload_result')
    payload = {}
    payload['variable'] = auto_var.parameter_id

    fp = tempfile.NamedTemporaryFile('w', delete=False)
    json.dump(ret, fp)
    fp.close()
    files = {'file_field': open(fp.name, 'r')}

    res = requests.post(url, data=payload, files=files)
    print(res.content)

def submit_parameter(auto_var):
    if auto_var.settings['server_url'] is None:
        raise ValueError
    url = urllib.parse.urljoin(auto_var.settings['server_url'], 'submit_parameter')
    payload = copy.deepcopy(auto_var.var_value)
    #payload['unique_name'] = auto_var.generate_name()
    res = requests.post(url, data=json.dumps(payload))
    response_content = json.loads(res.content)
    if response_content['has_result']:
        raise ParameterAlreadyRanError
    auto_var.parameter_id = response_content['id']
    return res



import argparse
from argparse import RawTextHelpFormatter
import base64
from copy import deepcopy
import inspect
import time
from typing import Dict, Tuple, List, Any, Callable, Optional, Union
import re
import pprint
import logging
import gc
import os

import git
import git.exc
import joblib
from joblib import Parallel, delayed
from mkdir_p import mkdir_p
from sklearn.model_selection import ParameterGrid

from .base import default_fn_dict, default_val_dict, \
        ParameterAlreadyRanError, VariableValueNotSetError, \
        VariableNotRegisteredError

logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s',
    level=logging.WARNING,
    datefmt='%Y-%m-%d %H:%M:%S')
logger = logging.getLogger(__name__)


def add_all_commit(repo, commit_msg="update"):
    repo.git.commit('-m', commit_msg)
    files = repo.git.diff(None, name_only=True)
    for f in files.split('\n'):
        repo.git.add(f)

class AutoVar(object):

    def __init__(self, before_experiment_hooks=None, after_experiment_hooks=None,
                 settings: Dict=None, logging_level: int=logging.WARNING) -> None:
        """
        settings : {
            'server_url': 'http://127.0.0.1:8080/nn_attack/',
            'result_file_dir': './results/'
            'file_format': 'json' or 'pickle'
        }
        """
        logger.setLevel(logging_level)

        self.experiments: Dict[str, Any] = {}
        self.variables: Dict[str, dict] = {}
        self.var_shown_name: Dict[str, Dict[str, str]] = {}
        self.var_description: Dict[str, str] = {}
        self.var_class: Dict[str, Any] = {}
        self.var_value: Dict[str, Any] = {}

        self.inter_var: Dict[str, Any] = {}
        self.result_fields: List[str] = []
        self.settings: Dict
        self.settings = {'file_format': 'json'}
        if isinstance(settings, dict):
            self.settings.update(settings)
        if ('result_file_dir' in self.settings) and self.settings['result_file_dir']:
            mkdir_p(self.settings['result_file_dir'])

        try:
            self.repo = git.Repo(search_parent_directories=True)
            self.var_value['git_hash'] = self.repo.head.object.hexsha
        except git.exc.InvalidGitRepositoryError:
            logger.warning("Git repository not found.")

        self.after_experiment_hooks = after_experiment_hooks
        self.before_experiment_hooks = before_experiment_hooks

        self._read_only: bool = False
        self._no_hooks: bool = False

    def set_logging_level(self, level: int):
        logger.setLevel(level)

    def add_variable_class(self, variable_class, var_name: str = None):
        if var_name is None:
            var_name = variable_class.__class__.var_name
            if var_name is None:
                var_name = type(variable_class).__name__

        if var_name in self.var_class:
            raise ValueError('Don\'t register class "{}" twice '.format(var_name))
        self.var_class[var_name] = variable_class
        #self.variables.setdefault(var_name, deepcopy(default_fn_dict))
        self.variables.update(variable_class.__class__.variables)
        self.var_shown_name.update(
                variable_class.__class__.variable_shown_name)

        if variable_class.default is not None:
            self._check_var_argument(var_name=var_name, argument=variable_class.default)

    def add_variable(self, var_name: str, dtype, default=None) -> None:
        d = deepcopy(default_val_dict)
        d['dtype'] = dtype
        self.variables.setdefault(var_name, d)
        if default is not None:
            self._check_var_argument(var_name=var_name, argument=default)
            d['default'] = default

    def get_var_shown_name(self, var_name: str,
                           argument: Union[str, None] = None) -> str:
        if argument is None:
            argument = self.get_variable_name(var_name)
        for key, val in self.var_shown_name[var_name].items():
            if re.fullmatch(key, argument):
                argument = re.sub(key, val, argument)
        return argument

    def get_var(self, var_name: str, *args, **kwargs):
        """[summary]

        Arguments:
            var_name {[type]} -- [description]
            argument {[type]} -- [description]

        Raises:
            ValueError -- [description]
            ValueError -- [description]

        Returns:
            [type] -- [description]
        """
        if var_name not in self.variables:
            raise VariableNotRegisteredError('Variable "%s" not registered.' % var_name)
        if var_name not in self.var_value:
            if self.variables[var_name]['type'] == 'choice':
                if self.var_class[var_name].default is not None:
                    argument = self.var_class[var_name].default
                else:
                    raise VariableValueNotSetError('Value for variable "%s" is not assigned.' % var_name)
            elif self.variables[var_name]['default'] is not None:
                argument = self.variables[var_name]['default']
            else:
                raise VariableValueNotSetError('Value for variable "%s" is not assigned.' % var_name)
        else:
            argument = self.var_value[var_name]

        return self.get_var_with_argument(
                var_name=var_name, argument=argument, *args, **kwargs)


    def match_variable(self, var_name: str, argument):
        if argument is None:
            return False

        if self.variables[var_name]["type"] == "val":
            return True
        else:
            if argument in self.variables[var_name]["argument_fn"]:
                return True
            else:
                for arg_template in self.variables[var_name]["argument_fn"]:
                    m = re.fullmatch(arg_template, argument)
                    if m is not None:
                        return True
        return None

    def get_variable_value(self, var_name: str):
        logger.warn("Use get_variable_name instead.")
        if var_name not in self.var_value:
            raise ValueError(f"{var_name} not in var_value")
        return self.var_value[var_name]

    def get_variable_name(self, var_name: str):
        if var_name not in self.var_value:
            raise ValueError(f"{var_name} not in variable name")
        return self.var_value[var_name]

    def get_var_with_argument(self, var_name: str, argument: str, *args, **kwargs):
        if self.variables[var_name]["type"] == "val":
            return argument
        else:
            cache_dir = None
            if argument in self.variables[var_name]["argument_fn"]:
                func = self.variables[var_name]["argument_fn"][argument]
                cache_dir = self.variables[var_name]['cache_dirs'][argument]
                required_vars = self.variables[var_name]['required_vars'][argument]
            else:
                m = None
                for arg_template, func in self.variables[var_name]["argument_fn"].items():
                    m = re.fullmatch(arg_template, argument)
                    if m is not None:
                        kwargs.update(m.groupdict())
                        cache_dir = self.variables[var_name]['cache_dirs'][arg_template]
                        required_vars = self.variables[var_name]['required_vars'][arg_template]
                        break
                if m is None:
                    raise ValueError('Argument "%s" not matched in Variable '
                                     '"%s".' % (argument, var_name))

            kwargs['auto_var'] = self
            named_args = inspect.getfullargspec(func)[0]
            if 'var_value' in named_args:
                kwargs['var_value'] = self.var_value
            if 'inter_var' in named_args:
                kwargs['inter_var'] = self.inter_var

            if cache_dir is not None:
                mkdir_p(cache_dir)
                #var_used = {var_name: self.var_value[var_name]}
                var_used = {var_name: argument}
                if required_vars is not None:
                    for var in required_vars:
                        if var not in self.var_value:
                            raise ValueError('Variable "%s" required by Variable '
                                             '"%s" is not set.' % (var, argument))

                        var_used[var] = self.var_value[var]
                required_variables = [str(v) for k, v in sorted(var_used.items())]
                cache_filename = os.path.join(cache_dir, '-'.join(required_variables) + '.pkl')

                if os.path.exists(cache_filename):
                    try:
                        func_outputs = joblib.load(cache_filename)
                        logger.info(f"using result from cache file {cache_filename} ...")
                    except:
                        os.unlink(cache_filename)
                        func_outputs = func(*args, **kwargs)
                        logger.info(f"dumping cache file to {cache_filename} ...")
                        joblib.dump(func_outputs, cache_filename)
                else:
                    func_outputs = func(*args, **kwargs)
                    logger.info(f"dumping cache file to {cache_filename} ...")
                    joblib.dump(func_outputs, cache_filename)
            else:
                func_outputs = func(*args, **kwargs)

            return func_outputs

    def get_intermidiate_variable(self, var_name: str):
        return self.inter_var[var_name]

    def set_variable_value(self, var_name: str, value) -> None:
        if self._read_only:
            raise ValueError("should not set variable value while read_only")
        #if var_name not in self.var_value:
        #    raise ValueError(f"{var_name} not in var_value")
        self.var_value[var_name] = value
        self._check_var_argument(var_name=var_name, argument=value)

    def set_intermidiate_variable(self, var_name: str, value) -> None:
        self.inter_var[var_name] = value

    def set_variable_value_by_dict(self, var_value: Dict[str, Any]) -> None:
        if self._read_only:
            raise ValueError("should not set variable value while read_only")
        for k, v in var_value.items():
            self.set_variable_value(k, v)

    def _run_before_hooks(self):
        if not self._no_hooks and self.before_experiment_hooks is not None:
            for hook_fn in self.before_experiment_hooks:
                try:
                    hook_fn(self)
                except ParameterAlreadyRanError:
                    logger.warning("The result for this parameter is "
                                        "already ran.")
                    return False
                except:
                    print("Error occur during running before hooks.")
                    raise
        return True

    def _run_after_hooks(self, ret):
        if not self._no_hooks and self.after_experiment_hooks is not None:
            for hook_fn in self.after_experiment_hooks:
                hook_fn(self, ret)

    def run_single_experiment(self, experiment_fn: Union[Callable[..., Any], str],
                              with_hook: bool=True,
                              verbose: int=0) -> Optional[bool]:
        self._read_only = True
        if isinstance(experiment_fn, str):
            original_settings = deepcopy(self.settings)
            self.settings.update(self.experiments[experiment_fn]['settings'])

        ret = None
        ret_hook = True
        if with_hook:
            ret_hook = self._run_before_hooks()

        try:
            if ret_hook:
                start_time = time.time()
                if isinstance(experiment_fn, str):
                    ret = self.experiments[experiment_fn]['fn'](self)
                else:
                    ret = experiment_fn(self)
                end_time = time.time()
                logger.info("Running time: %f", end_time - start_time)
                if isinstance(ret, dict):
                    ret['running_time'] = end_time - start_time
                else:
                    # return value is not dict
                    pass
                ret['var_value'] = deepcopy(self.var_value)
        finally:
            if with_hook and ret_hook:
                self._run_after_hooks(ret)
            self.inter_var.clear()
            self._read_only = False
            gc.collect()

        if isinstance(experiment_fn, str):
            self.settings = original_settings
        return ret

    def _check_var_argument(self, var_name: str, argument):
        if self.variables[var_name]["type"] != "val" \
            and not self.match_variable(var_name, argument):
                raise ValueError('Argument "%s" not in Variable "%s".' % (argument, var_name))
        elif self.variables[var_name]["type"] == "val" \
            and not isinstance(argument, self.variables[var_name]['dtype']):
            raise TypeError('Argument "%s" not in the correct type "%s" for Variable "%s".' % (
                            type(argument), self.variables[var_name]['dtype'], var_name))

    def _check_grid_params(self, grid_params: Dict[str, List]) -> bool:
        # TODO check duplicated params
        # TODO duplicated check var argument since set_param_value will check
        for k, v in grid_params.items():
            for i in v:
                self._check_var_argument(k, i)
        return True

    def run_grid_params(self,
                        experiment_fn: Union[Callable[..., Any], str],
                        grid_params: Union[Dict[str, List], List[Dict[str, List]]],
                        with_hook: bool=True,
                        max_params: int=-1,
                        commit_before_run: bool=False,
                        verbose: int=0,
                        allow_failure: bool=True,
                        n_jobs: int=-1,
                        backend: Optional[str]=None,
                        pre_dispatch: str='2 * n_jobs') -> Tuple[List, List]:

        if commit_before_run:
            if self.repo is None:
                raise ValueError("Not currently in git repo.")
            add_all_commit(self.repo)

        if isinstance(grid_params, list):
            ret_params: List = []
            for grid_param in grid_params:
                self._check_grid_params(grid_param)

                grid = ParameterGrid(grid_param)
                ret_params += list(grid)
        else:
            ret_params = list(ParameterGrid(grid_params))

        if max_params != -1:
            ret_params = ret_params[:max_params]

        parallel = Parallel(n_jobs=n_jobs, verbose=verbose,
                            backend=backend, pre_dispatch=pre_dispatch)

        def _helper(auto_var, params):
            auto_var.set_variable_value_by_dict(params)
            if verbose:
                logger.info("Running parameter:" + str(params))
            try:
                results = auto_var.run_single_experiment(
                    experiment_fn,
                    with_hook=(with_hook and (not self._no_hooks)),
                    verbose=verbose
                )
            except Exception as e:
                if allow_failure:
                    logger.error("Error with " + str(params))
                else:
                    raise e
                results = None

            return results

        with parallel:
            ret_results = parallel(delayed(_helper)(deepcopy(self), params)
                                   for params in ret_params)

        return ret_params, ret_results

    def summary(self) -> None:
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(self.variables)

    def generate_name(self) -> str:
        _ = hash(tuple(sorted(self.var_value.items())))
        return base64.b64encode(str(_).encode()).decode()

    def register_experiment(self, experiment_name: str,
                               experiment_fn: Callable[..., Any],
                               settings: Dict = None) -> None:
        if experiment_name in self.experiments:
            logger.info(f"Overriding experiment {experiment_name}")
        if settings is None:
            settings = dict()
        self.experiments[experiment_name] = {
            "fn": experiment_fn,
            "settings": settings,
        }

    def get_argparser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter)
        parser.add_argument('--no-hooks', required=False, action='store_true',
                            help="run without the hooks")

        if len(self.experiments) > 0:
            parser.add_argument('--experiment', type=str, required=True,
                                choices=[k for k in self.experiments.keys()],
                                help="The experiment to run")

        for var_name, v in self.variables.items():
            if v['type'] == 'choice':
                default = self.var_class[var_name].default
                required = False if default is not None else True
                help_str = ''
                if self.var_class[var_name].__doc__ is not None:
                    help_str += self.var_class[var_name].__doc__ + "\n"
                help_str += f"Options: (default: {default})\n"
                help_str += "\n".join(
                    ["  %s: %s" % (kk, vv.__doc__) for kk, vv in v['argument_fn'].items()])
                parser.add_argument(f'--{var_name}', type=str, required=required,
                                    action=make_action(self, var_name), default=default,
                                    help=help_str)
            else:
                default = v['default']
                required = False if default is not None else True
                parser.add_argument(f'--{var_name}', type=v['dtype'], required=required, default=default)

        return parser

    def parse_argparse(self, args: Optional[List[str]]=None):
        parser = self.get_argparser()
        parsed_args = parser.parse_args(args=args)
        self._no_hooks = parsed_args.no_hooks

        variables = vars(parsed_args)
        del variables['no_hooks']
        experiment_name = variables.pop('experiment', None)
        self.set_variable_value_by_dict(variables)

        if experiment_name:
            ret = self.run_single_experiment(experiment_name, with_hook=(not self._no_hooks))
            return ret

        return None



def make_action(auto_var, var_name):
    class CheckArgParseAction(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            argument = values
            if not auto_var.match_variable(var_name, argument):
                raise argparse.ArgumentTypeError(
                    'Argument "%s" not in Variable "%s".' % (argument, var_name))
            setattr(args, self.dest, values)
    return CheckArgParseAction



from typing import Any, Callable, Dict, List, Optional, Union
from ..auto_var import AutoVar

class Experiments():
    name: str
    experiment_fn: Callable[[AutoVar], Any]
    grid_params: Union[List[Dict[str, str]], Dict[str, str]]
    run_param: Dict[str, Any]

    def __init__(self):
        pass

    def __call__(self):
        return self.experiment_fn, self.name, self.grid_params, self.run_param



from functools import partial
import json
import os
import unittest
import shutil

import joblib
from numpy.testing import assert_array_equal
from mkdir_p import mkdir_p

from autovar import AutoVar
from autovar.base import RegisteringChoiceType, VariableClass, \
    register_var, VariableNotRegisteredError, VariableValueNotSetError
from autovar.hooks import save_result_to_file, default_get_file_name

class OrdVarClass(VariableClass, metaclass=RegisteringChoiceType):
    var_name = "ord"

    @register_var(argument='2')
    @staticmethod
    def l2(auto_var):
        return 2

    @register_var(argument='1')
    @staticmethod
    def l1(auto_var, var_value, inter_var):
        return 1

class TestAutovar(unittest.TestCase):
    def setUp(self):
        pass

    def test_json_file(self):
        settings = {'file_format': 'json', 'result_file_dir': 'test'}
        mkdir_p(settings['result_file_dir'])
        auto_var = AutoVar(
            settings=settings,
            after_experiment_hooks=[
                partial(save_result_to_file, get_name_fn=default_get_file_name)
            ],
        )
        auto_var.add_variable_class(OrdVarClass())
        auto_var.set_variable_value_by_dict({'ord': '1'})

        def experiment(auto_var):
            return {'test': auto_var.get_var('ord')}

        _ = auto_var.run_single_experiment(experiment, with_hook=True)

        with open("test/1.json", 'r') as f:
            ret = json.load(f)
        self.assertEqual(ret['test'], auto_var.get_var('ord'))
        shutil.rmtree(settings['result_file_dir'])



if __name__ == '__main__':
    unittest.main()



from typing import List, Optional
import joblib
import logging

_logger = logging.getLogger(__name__)

def register_var(var_type: str = 'fn_name', var_name: str = None,
                 argument: str = None, shown_name: str = None):
    """[summary]

    Arguments:
        func {[type]} -- [description]
        var_name {[type]} -- [description]
        argument {[type]} -- [description]
    """
    def decorator(func):
        data = {
            'var_type': var_type,
            'var_name': var_name,
            'argument': argument,
            'shown_name': shown_name,
        }
        if hasattr(func, 'registers'):
            func.registers.append(data)
        else:
            func.registers = [data]
        return func
    return decorator

def requires(required_vars: List[str]):
    """
    Should com after register_var decorator.
    """
    def decorator(func):
        if hasattr(func, 'registers'):
            for reg in func.registers:
                reg['required_vars'] = required_vars
        return func
    return decorator

def cache_outputs(cache_dir: str):
    """
    Should com after register_var decorator.
    """
    def decorator(func):
        if hasattr(func, 'registers'):
            for reg in func.registers:
                reg['cache_dir'] = cache_dir
        return func
    return decorator



#!/usr/bin/env python

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    install_requires = f.readlines()
install_requires = [line.strip() for line in install_requires]

setup(
    name='autovar',
    version='0.0.1a8',
    description='Experiment variables management framework.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/yangarbiter/autovar/',
    author='Yao-Yuan Yang',
    author_email='yangarbiter@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
    ],
    keywords='',
    packages=find_packages(exclude=[]),
    install_requires=install_requires,
    test_suite='autovar',
    extras_require={
        'dev': ['pylint'],
        'test': ['coverage', 'pylint', 'mypy'],
    },
    package_data={},
    entry_points={},
    project_urls={},
)



import argparse
import logging
import os.path
import tempfile
import unittest

from numpy.testing import assert_array_equal
from sklearn.datasets import make_moons
import joblib

from autovar import AutoVar
from autovar.base.decorators import cache_outputs, requires
from autovar.base import RegisteringChoiceType, VariableClass, \
    register_var, VariableNotRegisteredError, VariableValueNotSetError

class DatasetVarClass(VariableClass, metaclass=RegisteringChoiceType):
    """Dataset variable class"""
    var_name = 'dataset'
    default = None

    @register_var(argument=r"moon_(?P<n_samples>\d+)", shown_name="shown_halfmoon")
    @register_var(argument=r"halfmoon_(?P<n_samples>\d+)", shown_name="shown_halfmoon")
    @staticmethod
    def halfmoon(auto_var, var_value, n_samples):
        """halfmoon dataset"""
        X, y = make_moons(
            n_samples=int(n_samples),
            noise=0.3,
            random_state=1126,
        )
        return X, y

    @register_var(argument=r"no2_halfmoon_(?P<n_samples>\d+)", shown_name="no2_halfmoon")
    @staticmethod
    def no2_halfmoon(auto_var, var_value, n_samples):
        """no2 halfmoon dataset"""
        X, y = make_moons(
            n_samples=int(n_samples),
            noise=0.2,
            random_state=1126,
        )
        return X, y

    @cache_outputs(cache_dir=tempfile.TemporaryDirectory().name)
    @requires(['ord'])
    @register_var(argument=r"no4_halfmoon_(?P<n_samples>\d+)", shown_name="no4_halfmoon")
    @staticmethod
    def no4_halfmoon(auto_var, var_value, n_samples):
        """no2 halfmoon dataset"""
        X, y = make_moons(
            n_samples=int(n_samples),
            noise=0.4,
            random_state=1126,
        )
        return X, y

class OrdVarClass(VariableClass, metaclass=RegisteringChoiceType):
    var_name = "ord"
    default = "2"

    @register_var(argument='2')
    @staticmethod
    def l2(auto_var):
        return 2

    @register_var(argument='1')
    @staticmethod
    def l1(auto_var, var_value, inter_var):
        return 1

class TestAutovar(unittest.TestCase):

    def setUp(self):
        pass

    def test_cmd_args(self):
        auto_var = AutoVar(logging_level=logging.INFO)
        self.assertFalse(auto_var._no_hooks)
        args = ['--no-hooks']
        auto_var.parse_argparse(args=args)
        self.assertTrue(auto_var._no_hooks)

    def test_cmd_args_with_vars(self):
        auto_var = AutoVar(logging_level=logging.INFO)
        auto_var.add_variable_class(DatasetVarClass())

        args = ['--dataset', 'halfmoon_300']
        auto_var.parse_argparse(args=args)

        with self.assertRaises(argparse.ArgumentTypeError):
            args = ['--dataset', 'halfmaan_300']
            auto_var.parse_argparse(args=args)

        with self.assertRaises(SystemExit) as cm:
            args = []
            auto_var.parse_argparse(args=args)
        self.assertEqual(cm.exception.code, 2)

    def test_cmd_args_with_vars_2(self):
        auto_var = AutoVar(logging_level=logging.INFO)
        auto_var.add_variable_class(OrdVarClass())

        auto_var.parse_argparse(args=[])
        self.assertEqual(auto_var.get_var("ord"), 2)

    def test_cache_files(self):
        auto_var = AutoVar(logging_level=logging.INFO)
        auto_var.add_variable_class(DatasetVarClass())
        auto_var.add_variable_class(OrdVarClass())

        auto_var.set_variable_value("dataset", "no4_halfmoon_5")
        auto_var.set_variable_value("ord", "1")
        X, y = auto_var.get_var("dataset")
        cacheX, cachey = auto_var.get_var("dataset")

        assert_array_equal(X, cacheX)
        assert_array_equal(y, cachey)

        temp_dir = auto_var.variables['dataset']['cache_dirs']['no4_halfmoon_(?P<n_samples>\\d+)']

        cacheX, cachey = joblib.load(os.path.join(temp_dir, "no4_halfmoon_5-1.pkl"))
        assert_array_equal(X, cacheX)
        assert_array_equal(y, cachey)

    def test_val(self):
        auto_var = AutoVar(logging_level=logging.INFO)
        with self.assertRaises(VariableNotRegisteredError):
            auto_var.get_var('ord')

        auto_var.add_variable_class(OrdVarClass())
        auto_var.add_variable_class(DatasetVarClass())
        auto_var.add_variable('random_seed', int)

        with self.assertRaises(VariableValueNotSetError):
            auto_var.get_var('ord')

        auto_var.set_variable_value_by_dict({
            'ord': '1', 'dataset': 'halfmoon_200', 'random_seed': 1126
        })

        self.assertEqual(auto_var.get_var('ord'), 1)
        self.assertEqual(auto_var.get_var('random_seed'), 1126)
        self.assertEqual(len(auto_var.get_var('dataset')[0]), 200)
        self.assertEqual(
            len(auto_var.get_var_with_argument('dataset', 'halfmoon_300')[0]), 300)

        with self.assertRaises(ValueError):
            auto_var.set_variable_value_by_dict({'ord': 'l2'})


        with self.assertRaises(TypeError):
            auto_var.set_variable_value_by_dict({'random_seed': '1126.0'})

        self.assertEqual(
            auto_var.get_var_shown_name(var_name="dataset"),
            'shown_halfmoon')

        assert_array_equal(
            auto_var.get_var_with_argument('dataset', 'halfmoon_300')[0],
            auto_var.get_var_with_argument('dataset', 'moon_300')[0],
        )
        argparse_help = auto_var.get_argparser().format_help()
        self.assertTrue('halfmoon dataset' in argparse_help)
        self.assertTrue('Dataset variable class' in argparse_help)

    def test_run_grid(self):
        auto_var = AutoVar()
        auto_var.add_variable_class(OrdVarClass())
        auto_var.add_variable_class(DatasetVarClass())
        auto_var.add_variable('random_seed', int)

        grid_params = {
            "ord": ['1', '2'],
            "dataset": ['halfmoon_50', 'halfmoon_10'],
            "random_seed": [1126],
        }
        def fn(auto_var):
            return {
                "ord": auto_var.var_value['ord'],
                "dataset": auto_var.var_value['dataset'],
                "random_seed": auto_var.var_value['random_seed'],
            }
        params, results = auto_var.run_grid_params(
                fn, grid_params=grid_params, n_jobs=1)

        self.assertEqual(
            params[0],
            {'ord': '1', 'dataset': 'halfmoon_50', 'random_seed': 1126,})

        del results[0]['var_value']['git_hash']
        self.assertEqual(params[0], results[0]['var_value'])

        del results[0]['running_time']
        del results[0]['var_value']
        self.assertEqual(params[0], results[0])


if __name__ == '__main__':
    unittest.main()
